import logging
import os
import time
from datetime import datetime

from requests import Session, ConnectionError
from requests.structures import CaseInsensitiveDict

BASE_URL = "https://api.planfact.io"
MAX_SENDING_TRIES = 3
MIN_TIME_BETWEEN_REQUEST = 0.1
MAX_REQUESTS_PER_FUNCTION = 100

logger = logging.getLogger(__name__)


class PfInteraction:
    def __init__(self):
        self.session = Session()
        self.headers = CaseInsensitiveDict()
        self.headers["Accept"] = "application/json"
        self.headers["Content-Type"] = "application/json"
        self.time_at_last_request = time.time()

    def _get_header(self):
        PF_API_KEY = os.environ["PF_API_KEY"]
        self.headers["X-ApiKey"] = PF_API_KEY
        return self.headers

    def request_list(self, method, path, changes_from_date=None, data=None, params=None, headers=None):
        if params is not None:
            if method == 'get' and 'paging.limit' in params.keys():
                if changes_from_date is not None:
                    if not isinstance(changes_from_date, datetime):
                        raise TypeError('changes_from_date have to be from type datetime')
                    else:
                        params.update({
                            'sorting.field': ' ModifyDate',
                            'filter.changesFromDate': changes_from_date.strftime('%Y-%m-%dT%H:%M:%S')
                        })
            all_items = []
            for i in range(MAX_REQUESTS_PER_FUNCTION):
                res, status_code = self.request(method=method, path=path, data=data, params=params, headers=headers)
                all_items += res.json()['data']['items']
                if len(res.json()['data']['items']) < params['paging.limit']:
                    break
                else:
                    params['paging.offset'] = params['paging.limit'] * (i + 1)
            for item in all_items:
                # convert ModifyDate into datetime
                if 'modifyDate' in item.keys():
                    try:
                        item['modifyDate'] = datetime.strptime(item['modifyDate'], '%Y-%m-%dT%H:%M:%S.%f')
                    except ValueError:
                        item['modifyDate'] = datetime.strptime(item['modifyDate'], '%Y-%m-%dT%H:%M:%S')
                else:
                    break
            return all_items
        else:
            raise RuntimeError('this is not an list operation')

    def request(self, method, path, data=None, json: str = None, params=None, headers=None):
        if headers is None:
            headers = self._get_header()
        # make sure that the rate is not to high
        time_passed = time.time() - self.time_at_last_request
        logger.debug('time passed %s' % str(time_passed))
        # send the request
        for send_request_tries in range(MAX_SENDING_TRIES):
            # wait MIN_TIME_BETWEEN_REQUEST second
            if time_passed < MIN_TIME_BETWEEN_REQUEST:
                logger.debug('have to wait %s' % str(MIN_TIME_BETWEEN_REQUEST - time_passed))
                time.sleep(MIN_TIME_BETWEEN_REQUEST - time_passed)
            # reset time
            self.time_at_last_request = time.time()
            try:
                response = self.session.request(method,
                                                url=BASE_URL + path,
                                                data=json,
                                                json=data,
                                                params=params,
                                                headers=headers)
                logger.debug('sending pf-request')
            except ConnectionError as err:
                error = ('Got an exception while %s sending: ' % method) + str(err)
                logger.warning(error)
                logger.warning(f'current try: {send_request_tries}')
                if send_request_tries < 3:
                    send_request_tries += 1
                else:
                    raise ConnectionError(err.args[0].args[0])  # Sometimes Connection aborted.
            except Exception as err:
                error = ('Got an exception while %s sending: ' % method) + str(err)
                logger.error(error)
                raise
            else:
                if response.status_code == 200:
                    if response.json()['isSuccess']:
                        return response, 200
                    else:
                        raise ConnectionError(
                            "Status is 200, but no Success, with "
                            "errorCode: {} and errorMessage: '{}'".format(response.json()['errorCode'],
                                                                          response.json()['errorMessage']))
                else:
                    raise ConnectionError("Status not 200, but: {} ({})".format(response.status_code, response.text))


pf_interaction = PfInteraction()



import logging

from .models import *
from .pf_interaction import pf_interaction

logger = logging.getLogger(__name__)

PAGE_LIMIT = 500


def get_currencies():
    path = f'/api/v1/currencies'
    params = {'paging.limit': PAGE_LIMIT,
              'paging.offset': 0}
    items = pf_interaction.request_list('get', path, params=params)
    return items


def get_operations(account_id=None):
    path = f'/api/v1/operations'
    params = {'paging.limit': PAGE_LIMIT,
              'paging.offset': 0}
    if account_id is not None:
        params['filter.accountId'] = account_id
    items = pf_interaction.request_list('get', path, params=params)
    return items


def get_account_by_id(account_id: int):
    path = f"/api/v1/accounts/{account_id}"
    resp, _ = pf_interaction.request('get', path)
    return resp.json()['data']


def get_accounts(changes_from_date=None, add_params=None):
    path = f'/api/v1/accounts'
    params = {'paging.limit': PAGE_LIMIT}
    if add_params is not None:
        params.update(add_params)
    items = pf_interaction.request_list('get', path, changes_from_date, params=params)
    return items


def get_operationcategories(return_tree=False, tree_parent_ids=None):
    # GET /api/v1/operationcategories
    path = f'/api/v1/operationcategories'
    params = {'paging.limit': PAGE_LIMIT}
    items = pf_interaction.request_list('get', path, params=params)
    # only Income and Outcome categories
    items = [item for item in items if item['operationCategoryType'] in ['Income', 'Outcome']]
    items = sorted(items, key=lambda cat_type: cat_type['operationCategoryType'])
    # check if item got children:
    parent_ids = [item['parentOperationCategoryId'] for item in items if item['parentOperationCategoryId'] is not None]
    for item in items:
        if item['operationCategoryId'] in parent_ids:
            item['is_parent'] = True
        else:
            item['is_parent'] = False
    if not return_tree:
        # return only non parent categories
        return [item for item in items if item['is_parent'] is False]
    else:
        # build tree
        if tree_parent_ids is None:
            parents = [item for item in items if item['parentOperationCategoryId'] is None]
        else:
            parents = [item for item in items if item['operationCategoryId'] in tree_parent_ids]
        level = 0
        tree_items = list()
        while True:
            children = list()
            tree_items += parents
            for parent in parents:
                parent['level'] = level
                tmp_children = [item for item in items
                                if item['parentOperationCategoryId'] == parent['operationCategoryId']]
                parent['childsOperationCategoryIds'] = [item['operationCategoryId'] for item in tmp_children]
                children += tmp_children
            parents = children
            level += 1
            if not parents:
                break
        return tree_items


def get_projects_groups(changes_from_date=None):
    path = f'/api/v1/projects/groups'
    params = {'paging.limit': PAGE_LIMIT}
    items = pf_interaction.request_list('get', path, changes_from_date, params=params)
    return items


def get_projects(changes_from_date=None):
    path = f'/api/v1/projects'
    params = {'paging.limit': PAGE_LIMIT}
    items = pf_interaction.request_list('get', path, changes_from_date, params=params)
    return items


def create_income(operation: IncomeOutcomeOperation) -> OperationModel:
    path = f'/api/v1/operations/income'
    json = operation.json()
    res, _ = pf_interaction.request('post', path, json=json)
    return OperationModel.parse_obj(res.json()['data'])


def create_outcome(operation: IncomeOutcomeOperation) -> OperationModel:
    path = f'/api/v1/operations/outcome'
    json = operation.json()
    res, _ = pf_interaction.request('post', path, json=json)
    return OperationModel.parse_obj(res.json()['data'])


def create_move_operation(operation: MoveOperation) -> List[OperationModel]:
    path = f'/api/v1/operations/move'
    json = operation.json()
    res, _ = pf_interaction.request('post', path, json=json)
    return [OperationModel.parse_obj(d) for d in res.json()['data']]


def create_account(account: Account):
    path = f'/api/v1/accounts'
    data = account.dict()
    res, _ = pf_interaction.request('post', path, data=data)
    return res.json()['data']


def get_allowed_entities(rule, entities, ident_by):
    allowed_ids = None
    if rule['accessRuleType'] == 'Allowed':
        allowed_ids = rule['ids']
    elif rule['accessRuleType'] == 'Disallowed':
        available_ids = [item[ident_by] for item in entities]
        allowed_ids = list(set(available_ids) - set(rule['ids']))
    allowed_entities = [item for item in entities if item[ident_by] in allowed_ids]
    return allowed_entities


def get_user_permissions(user_id, opcat_parent_ids=None):
    path = f'/api/v1/accesscontexts'
    params = {'filter.kind': 'ByBusiness',
              'paging.limit': PAGE_LIMIT}
    items = pf_interaction.request_list('get', path, params=params)
    user = next((item for item in items if item['user']['id'] == user_id), None)
    # get user permission for accounts, projects and categories
    permissions = {}
    # accounts
    entities = get_accounts()
    allowed_entities = get_allowed_entities(user['accountsRule'], entities, 'accountId')
    permissions['accounts'] = allowed_entities
    # categories
    entities = get_operationcategories(return_tree=True, tree_parent_ids=opcat_parent_ids)
    allowed_entities = get_allowed_entities(user['categoriesRule'], entities, 'operationCategoryId')
    permissions['categories'] = allowed_entities
    # projects
    entities = get_projects()
    allowed_entities = get_allowed_entities(user['projectsRule'], entities, 'projectId')
    permissions['projects'] = allowed_entities
    entities = get_projects_groups()
    allowed_entities = get_allowed_entities(user['projectGroupsRule'], entities, 'projectGroupId')
    permissions['project_groups'] = allowed_entities

    return user['user'], permissions


def get_users():
    path = '/api/v1/accesscontexts'
    params = {'filter.kind': 'ByBusiness',
              'paging.limit': PAGE_LIMIT}
    items = pf_interaction.request_list('get', path, params=params)
    items = [item['user'] for item in items]
    return items


def get_account_history():
    path = '/api/v1/bizinfos/accountshistory'
    resp, _ = pf_interaction.request('get', path)
    return resp.json()["data"]


def get_balance_by_id(account_id: int):
    account = get_account_by_id(account_id)
    all_account_histories = get_account_history()
    account_hist = next((ah for ah in all_account_histories if ah['accountId'] == account['accountId']), None)
    if account_hist is None:
        balance = 0.
    else:
        fact_values = [float(fh["factValue"]) for fh in account_hist["details"]]
        if account["startingRemainderValue"] is None:
            account["startingRemainderValue"] = 0
        balance = sum(fact_values) + account["startingRemainderValue"]
    return balance


def get_balances():
    accounts = get_accounts()
    all_account_histories = get_account_history()
    balances = list()
    for account in accounts:
        account_hist = next((ah for ah in all_account_histories if ah['accountId'] == account['accountId']), None)
        if account_hist is None:
            balance = 0.
        else:
            fact_values = [float(fh["factValue"]) for fh in account_hist["details"]]
            if account["startingRemainderValue"] is None:
                account["startingRemainderValue"] = 0
            balance = sum(fact_values) + account["startingRemainderValue"]
        balances.append({'accountId': account['accountId'],
                         'title': account['title'],
                         'balance': balance})
    return balances



from datetime import datetime
from json import dumps
from typing import List, Optional

from pydantic import BaseModel, validator, NonNegativeFloat


def custom_dumps(args, *, default):
    for k, v in args.items():
        if isinstance(v, datetime):
            args[k] = v.strftime('%Y-%m-%d')
    return dumps(args, default=default)


class OperationItems(BaseModel):
    calculationDate: datetime
    isCalculationCommitted: bool
    contrAgentId: Optional[int]
    operationCategoryId: Optional[int]
    projectId: Optional[int]
    firstAdditionalOperationAttributeId: Optional[int]
    value: NonNegativeFloat

    class Config:
        json_dumps = custom_dumps

    @validator('projectId', 'operationCategoryId')
    def check_for_zero_ids(cls, v):
        if v == 0:
            return None
        else:
            return v


class IncomeOutcomeOperation(BaseModel):
    operationDate: datetime
    calculationDate: Optional[datetime]
    isCalculationCommitted: Optional[bool]
    contrAgentId: Optional[int]
    accountId: int
    operationCategoryId: Optional[int]
    comment: Optional[str]
    value: Optional[NonNegativeFloat]
    isCommitted: Optional[bool]
    items: Optional[List[OperationItems]]
    externalId: Optional[str]
    distributeCalculationDate: Optional[datetime]
    distributeCalculationType: Optional[str]

    class Config:
        json_dumps = custom_dumps


class MoveOperation(BaseModel):
    """
    valueByProjects: object     Сумма по проектам  not implemented!
    """
    debitingDate: datetime  # Дата списания
    admissionDate: datetime  # Дата зачисления
    debitingAccountId: int  # Счет списания
    admissionAccountId: int  # Счет зачисления
    debitingValue: Optional[NonNegativeFloat]  # Сумма списания
    admissionValue: Optional[NonNegativeFloat]  # Сумма зачисления
    comment: Optional[str]  # Комментарий
    isCommitted: Optional[bool]  # Признак того, что операция проведена
    importLogId: Optional[int]  # Идентификатор лога импорта
    debitingItems: Optional[List[OperationItems]]  # Части операции списания
    admissionItems: Optional[List[OperationItems]]  # Части операции зачисления
    debitingExternalId: Optional[str]  # Внешний идентификатор для списания
    admissionExternalId: Optional[str]  # Внешний идентификатор для зачисления

    class Config:
        json_dumps = custom_dumps

    def set_to_committed(self):
        self.isCommitted = True
        for item in self.debitingItems or []:
            item.isCalculationCommitted = True
            item.calculationDate = self.debitingDate
        for item in self.admissionItems or []:
            item.isCalculationCommitted = True
            item.calculationDate = self.admissionDate


class Account(BaseModel):
    title: str
    companyId: int
    accountType: str
    currencyCode: str
    longTitle: Optional[str]
    description: Optional[str]
    accountType: Optional[str]
    accountAcct: Optional[str]
    correspondentAcct: Optional[str]
    accountBik: Optional[str]
    accountBank: Optional[str]
    active: Optional[bool]
    remainder: Optional[float]
    remainderDate: datetime
    externalId: Optional[str]
    accountGroupId: Optional[int]


class AccountLink(BaseModel):
    accountId: int
    companyId: int
    externalId: Optional[str]
    title: Optional[str]


class CompanyLinkModel(BaseModel):
    companyId: int
    externalId: Optional[str]
    title: Optional[str]


class CurrencyLinkModel(BaseModel):
    currencyCode: str
    title: Optional[str]


class OperationCategoryLinkModel(BaseModel):
    operationCategoryId: int
    title: Optional[str]
    externalId: Optional[str]


class ProjectLinkModel(BaseModel):
    projectId: int
    title: Optional[str]
    externalId: Optional[str]


class ContrAgentLinkModel(BaseModel):
    contrAgentId: int
    title: Optional[str]
    externalId: Optional[str]


class OperationPartModel(BaseModel):
    isCalculationCommitted: bool
    project: Optional[ProjectLinkModel]
    contrAgent: Optional[ContrAgentLinkModel]
    operationCategory: Optional[OperationCategoryLinkModel]
    value: float
    valueInUserCurrency: float


class OperationModel(BaseModel):
    account: AccountLink
    accountCompany: CompanyLinkModel
    accountCurrency: CurrencyLinkModel
    boundMoveOperationId: Optional[int]
    comment: Optional[str]
    isCommitted: Optional[bool]
    operationCategory: OperationCategoryLinkModel
    operationId: int
    operationParts: List[OperationPartModel]
    operationType: str
    value: float
    valueInUserCurrency: float
    externalId: Optional[str]
